%{
#include "sds.h"
#include "structs.h"
#include "iascc.tab.h"

#define YY_USER_ACTION yylloc.first_line = yylloc.last_line = yylineno;

void error(int, const char*);
int fileno(FILE*);

/* flex specification for our language.

we count lines by keeping a counter and advancing it on \n.
two tokens are used for integers and labels so whitespace isn't
allowed in between.

a state is used to make handling of literal strings easier
*/
%}
%option nounput
%option noinput
%x string
%%
[ \t] ;
,                     { return ','; }
~                     { return '~'; }
\(                    { return '('; }
\)                    { return ')'; }
\n                    { yylineno++; return '\n'; }
\.data                { return DIRECTIVE_DATA; }
\.string              { return DIRECTIVE_STRING; }
\.entry               { return DIRECTIVE_ENTRY; }
\.extern              { return DIRECTIVE_EXTERN; }
[+-]?[0-9]+           { yylval.ival = atoi(yytext); return INT; }
#[+-]?[0-9]+          { yylval.ival = atoi(yytext+1); return OP_INT; }
\"                    { BEGIN string; yylval.sval = sdsempty(); }
<string>[^\\"\n]*     { yylval.sval = sdscat(yylval.sval, yytext); }
<string>\\n           { yylval.sval = sdscat(yylval.sval, "\n"); }
<string>\\t           { yylval.sval = sdscat(yylval.sval, "\t"); }
<string>\\[\\"]       { yylval.sval = sdscat(yylval.sval, yytext + 1); }
<string>\"            { BEGIN 0; return STRING; }
<string>\\.           { error(yylineno, "bogus escape in string"); BEGIN 0; }
<string>\n            { error(yylineno, "newline in quoted string");
                        yylineno++;
                        BEGIN 0; }
r[0-7]                { yylval.ival = atoi(yytext+1); return REGISTER; }
mov                   { yylval.ival = INST_MOV; return INST_MOV; }
cmp                   { yylval.ival = INST_CMP; return INST_CMP; }
add                   { yylval.ival = INST_ADD; return INST_ADD; }
sub                   { yylval.ival = INST_SUB; return INST_SUB; }
lea                   { yylval.ival = INST_LEA; return INST_LEA; }
not                   { yylval.ival = INST_NOT; return INST_NOT; }
clr                   { yylval.ival = INST_CLR; return INST_CLR; }
inc                   { yylval.ival = INST_INC; return INST_INC; }
dec                   { yylval.ival = INST_DEC; return INST_DEC; }
jmp                   { yylval.ival = INST_JMP; return INST_JMP; }
bne                   { yylval.ival = INST_BNE; return INST_BNE; }
red                   { yylval.ival = INST_RED; return INST_RED; }
prn                   { yylval.ival = INST_PRN; return INST_PRN; }
jsr                   { yylval.ival = INST_JSR; return INST_JSR; }
rts                   { yylval.ival = INST_RTS; return INST_RTS; }
stop                  { yylval.ival = INST_STOP; return INST_STOP; }
[a-zA-z][a-zA-Z0-9]*: { yylval.sval = sdsnew(yytext);
                        sdsrange(yylval.sval, 0, -2);
                        return LABEL; }
[a-zA-z][a-zA-Z0-9]*  { yylval.sval = sdsnew(yytext);
                        return OP_LABEL; }
;[^\r\n]*[\r\n]       { yylineno++; return COMMENT; }
<<EOF>>               { static int once = 0; return (once = !once) ? '\n' : 0; }
.
%%
#ifdef LEXER_MAIN

int main() {
    int token = yylex();

    while (token != 0)
        token = yylex();
}

#endif
